'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.META_SOURCE = exports.META_TIMESTAMP = exports.RESET_ACTION = exports.REWIND_ACTION = undefined;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _get = function get(object, property, receiver) { if (object === null) object = Function.prototype; var desc = Object.getOwnPropertyDescriptor(object, property); if (desc === undefined) { var parent = Object.getPrototypeOf(object); if (parent === null) { return undefined; } else { return get(parent, property, receiver); } } else if ("value" in desc) { return desc.value; } else { var getter = desc.get; if (getter === undefined) { return undefined; } return getter.call(receiver); } };

exports.isGossipType = isGossipType;

var _scuttlebutt = require('scuttlebutt');

var _scuttlebutt2 = _interopRequireDefault(_scuttlebutt);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

var REWIND_ACTION = exports.REWIND_ACTION = '@@scuttlebutt/REWIND';
var RESET_ACTION = exports.RESET_ACTION = '@@scuttlebutt/RESET';

var META_TIMESTAMP = exports.META_TIMESTAMP = '@@scuttlebutt/TIMESTAMP';
var META_SOURCE = exports.META_SOURCE = '@@scuttlebutt/SOURCE';

// update and state history structure keys
var UPDATE_ACTION = 0,
    UPDATE_TIMESTAMP = 1,
    UPDATE_SOURCE = 2,

// state snapshot is post-action
STATE_ACTION = 0,
    STATE_TIMESTAMP = 1,
    STATE_SNAPSHOT = 2;

// ignore actiontypes beginning with @
// by default just pass through missing types (redux will blow up later)
function isGossipType() {
  var type = arguments.length <= 0 || arguments[0] === undefined ? '' : arguments[0];

  return type.substr(0, 1) !== '@';
}

// not used any more?
var formatSource = function formatSource() {
  var source = arguments.length <= 0 || arguments[0] === undefined ? '????' : arguments[0];
  return source.substr(0, 2) + '…' + source.substr(-4, 4);
},
    formatTime = function formatTime(time) {
  return new Date(time).toJSON().substr(-10, 9);
},
    formatUpdate = function formatUpdate(action, full) {
  return formatSource(action[2]) + ' ' + formatTime(action[1]) + ' ' + (full ? JSON.stringify(action[0]) : '~');
};

var Dispatcher = function (_Scuttlebutt) {
  _inherits(Dispatcher, _Scuttlebutt);

  function Dispatcher() {
    _classCallCheck(this, Dispatcher);

    // history of all current updates
    // in-recieved-order is for scuttlebutt, sorted for time travel

    var _this = _possibleConstructorReturn(this, Object.getPrototypeOf(Dispatcher).call(this));

    _this._updates = [];
    _this._updateHistory = [];

    // history of store states after each action was applied
    _this._stateHistory = [];

    // dispatch to redux
    _this._reduxDispatch = function () {
      throw new Error('Are you sure you called wrapDispatch?');
    };
    return _this;
  }

  // wraps the redux dispatch


  _createClass(Dispatcher, [{
    key: 'wrapDispatch',
    value: function wrapDispatch(dispatch) {
      var _this2 = this;

      this._reduxDispatch = dispatch;

      return function (action) {
        // apply this action to our scuttlebutt model (and send to peers). It
        // will dispatch, taking care of the the appropriate time ordering
        if (isGossipType(action.type)) {
          _this2.localUpdate(action);
        } else {
          return dispatch(action);
        }
      };
    }

    // wraps the initial state, if any, into the first snapshot

  }, {
    key: 'wrapInitialState',
    value: function wrapInitialState(initialState) {
      return initialState && [,, initialState];
    }

    // rewinds history when it changes

  }, {
    key: 'wrapReducer',
    value: function wrapReducer(reducer) {
      // wrap the root reducer to track history and rewind occasionally
      // currentState is our higher-order form, an array of [snapshot, timestamp]
      return function () {
        var currentState = arguments.length <= 0 || arguments[0] === undefined ? [] : arguments[0];
        var action = arguments[1];

        var timestamp = action && action.meta && action.meta[META_TIMESTAMP];
        var stateIndex = void 0;

        // this will dispatch an action after the snapshot preceding it, then will
        // replay actions which occurred after it (if any)
        // rewind to -1 for "before the start of time"
        for (stateIndex = currentState.length - 1; stateIndex >= -1; stateIndex--) {
          var thisTimestamp = currentState[stateIndex] && currentState[stateIndex][STATE_TIMESTAMP],
              thisSnapshot = stateIndex === -1 ? undefined : currentState[stateIndex][STATE_SNAPSHOT];

          // thisTimestamp will be undefined until the first timestamped action.
          // if this action has no timestamp, we're before the start of time, or,
          // crucially, if this action is newer than the this snapshot
          if (!timestamp || !thisTimestamp || stateIndex === -1 || timestamp > thisTimestamp) {
            // add to history in the shape [ACTION, TIMESTAMP, SNAPSHOT]
            // splice doesn't perform well, btw.
            currentState.splice(stateIndex + 1, 0, [action, timestamp || thisTimestamp, // in case on missing timestamp
            reducer(thisSnapshot, action)]);

            // -1 for length to index, -1 for the additional element we just added to the array
            /*
            if (stateIndex !== currentState.length - 2 && typeof window !== 'undefined') {
              console.log(`time travelled ${timestamp} after ${thisTimestamp} (𝛥${timestamp - thisTimestamp}ms)`)
            }
            */

            break;
          }
        }

        // replay any actions after the one just dispatched. skip the last item
        // of the array, so it won't run for regular, in-order actions
        // skip the inserted action (index + 1)
        for (stateIndex = stateIndex + 2; stateIndex < currentState.length; stateIndex++) {
          var thisState = currentState[stateIndex],
              thisAction = thisState[STATE_ACTION],
              lastState = currentState[stateIndex - 1],
              lastSnapshot = lastState ? lastState[STATE_SNAPSHOT] : undefined;

          /*
          * FIXME: very destructive action. if the update fails, we'll have fucked
          * all the past histories... but surely they can be replayed.
          * applyUpdate() should dispatch a special "FORGET_ACTION" which removes
          * our timestamped new action and replays the previously-good state.
          * this realistically means we're **quietly catching all errors in the
          * reducer chain**.
          */

          // update each state snapshot, secretly hoping each action passes
          // console.log(`-> replaying action ${stateIndex}`,
          //   thisAction, thisState[STATE_TIMESTAMP], thisAction === action)

          thisState[STATE_SNAPSHOT] = reducer(lastSnapshot, thisAction);

          // deubg: add a checksum of the ordered timestamps to the store
          if (typeof window !== 'undefined') {
            var checksum = function checksum(arr) {
              var chk = 0x12345678;

              for (var i = 0; i < arr.length && !isNaN(arr[i]); i++) {
                chk += arr[i] * (i + 1);
              }

              return chk;
            };

            window.__checksum = checksum(currentState.map(function (state) {
              return state[STATE_TIMESTAMP];
            }));
          }
        }

        // if we're here, the currentState history has been updated
        return currentState;
      };
    }

    // wraps getState to return the latest state snapshot
    // will always be called after @@redux/INIT

  }, {
    key: 'wrapGetState',
    value: function wrapGetState(getState) {
      return function () {
        var state = getState(),
            lastState = state[state.length - 1];

        return lastState && lastState[STATE_SNAPSHOT];
      };
    }

    // Apply update (action) to our store
    // implemented for scuttlebutt class

  }, {
    key: 'applyUpdate',
    value: function applyUpdate(update) {
      var _extends2;

      var _update = _slicedToArray(update, 3);

      var action = _update[0];
      var timestamp = _update[1];
      var source = _update[2];
      // [, lastTimestamp] = this._updateHistory[this._updateHistory.length - 1] || [, 0],
      // add our metadata to the action
      var localAction = _extends({}, action, {
        meta: _extends({}, action.meta, (_extends2 = {}, _defineProperty(_extends2, META_TIMESTAMP, timestamp), _defineProperty(_extends2, META_SOURCE, source), _extends2))
      });

      // we log all updates to emit in the order we saw them.
      // not sure if it's better than replaying in order of timestamp (which might
      // cut down on the amount of time travelling done by all peers), but seems
      // like the de facto for scuttlebutt models
      this._updates.push(update);

      try {
        this._reduxDispatch(localAction);
      } catch (error) {
        // an update failed, so we'll consider this update invalid
        console.error('error applying update', localAction);
        console.error(error.stack);

        /*
        * TODO: recover from erroroneous action:
        *   remove update, replay old history. could do the actual splice
        *   after the update succeeds? better performance under conflict-
        *   prone conditions
        */
        if (typeof window !== 'undefined') console.warn('your store is probably in a terrible state rn');

        return false; // do not emit to peers
      }

      return true;

      console.info('applyUpdate success', update);
    }

    // gossip
    // implemented for scuttlebutt class

  }, {
    key: 'history',
    value: function history(sources) {
      return this._updates.filter(function (update) {
        return (0, _scuttlebutt.filter)(update, sources);
      });
    }

    // apply an update locally
    // we should ensure we don't send objects which will explode JSON.parse here
    // implemented over scuttlebutt class

  }, {
    key: 'localUpdate',
    value: function localUpdate(action) {
      if (process.env.NODE_ENV === 'development') {
        try {
          _get(Object.getPrototypeOf(Dispatcher.prototype), 'localUpdate', this).call(this, action);
        } catch (error) {
          throw new Error('Scuttlebutt couldn\'t dispatch', error);
        }
      } else {
        // try our luck
        _get(Object.getPrototypeOf(Dispatcher.prototype), 'localUpdate', this).call(this, action);
      }
    }

    // super.localUpdate(this._filterUpdate(action))
    // Recurse through the value and attempt to prune unserializable leaf objects.
    // A well-structured app won't be dispatching bad actions like this, so
    // this might become a dev-only check. also, it's far from foolproof.

  }, {
    key: '_filterUpdate',
    value: function _filterUpdate(value) {
      if ((typeof value === 'undefined' ? 'undefined' : _typeof(value)) !== 'object') return value;

      if (value && value.constructor && /(^Synthetic|Event$)/.test(value.constructor.name)) return null;

      var result = {};
      for (var prop in value) {
        result[prop] = this._filterUpdate(value[prop]);
      }
      return result;
    }
  }]);

  return Dispatcher;
}(_scuttlebutt2.default);

exports.default = Dispatcher;